// Lzsa1Decode:
#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define AX	%ax
#define BX	%bx
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define AX	%eax
#define BX	%ebx
#define SI	%esi
#define DI	%edi
#endif

MATCH_RUN_LEN		=	15
LITERALS_RUN_LEN	=	7
MIN_MATCH_SIZE		=	3
MIN_LITERALS_SIZE	=	0

#define PACKED_ONLY			// assume no copy block, optional
//#define PARANOIA			// cover rare cases, optional
 
.macro	shrclw cnt,obj
#ifdef ONLY8086
	movb	\cnt, %cl
	shrw	%cl, \obj
#else
	shrw	\cnt, \obj
#endif
.endm

#ifdef FLAT16OUT
#define RAW_FORMAT
#endif
lzsa1main:
#ifdef PARANOIA
	cld
#endif
#ifndef RAW_FORMAT
# ifndef NO_LZSA1_HEADER
	lodsw
	cmpw	$0x9E7B, %ax	// magic
	jne	lzsa1main
	lodsb
//	cmpb	$0, %al		// lzsa1
//	jne	lzsa1main
# endif
	xorw	%ax, %ax
	xchgw	%ax, %di
	shrclw	$4, %ax
	jmp	lzsa1blockz	// %di *MUST* be paragraph aligned
# ifndef PACKED_ONLY
lzsa1copy:
	movsb			// handle 64K case
	decw	%cx
	rep	movsb		// copy block
# endif
lzsa1block:			// uncompress chunk
	movw	$0x1000, %ax
lzsa1blockz:	
	movw	%es, %bx
	addw	%ax, %bx
	movw	%bx, %es
# ifndef FLAT16
	movw	%si, %ax
	andw	$0xf, %si
	shrclw	$4, %ax
	movw	%ds, %bx
	addw	%ax, %bx
	movw	%bx, %ds
# endif
	lodsw			// block size
	xchgw	%ax, %cx
	movw	%cx, %dx
	lodsb
# ifndef PACKED_ONLY
	orb	%al, %al
	js	lzsa1copy
	jne	lzsa1full	// 64Kb block
# endif
	jcxz	lzsa1quit	// bail if we hit EOD
lzsa1full:
	addw	%si, %dx
#endif
lzsa1chunk:			// uncompress chunk
	lodsb			// get token O|LLL|MMMM
	movb	%al, %bl	// keep token in bl
	shrclw	$4, %ax		// shift literals length into place
	movw	$LITERALS_RUN_LEN+256*MIN_LITERALS_SIZE, %cx
	call	lzsa1len	// %ch = LITERALS_RUN_LEN
	rep	movsb		// copy %cx literals from %ds:%si to %es:%di
#ifndef RAW_FORMAT
	cmpw	%dx, %si
	je	lzsa1block	// bail if we hit EOD
#endif
#ifdef FLAT32
	orl	$-1, %eax
#endif
	testb	%bl, %bl	// check match offset size in token (O bit)
	js	lzsa1LongOfs
#ifndef FLAT32
	movb	$-1, %ah	// set offset bits 15-8 to 1
#endif
	lodsb
	.byte	0x3C		// mask lodsw with cmpb $0xAD, %al
lzsa1LongOfs:
	lodsw
	xchg	AX, BX		// %bx: match offset  %ax: original token
	movw	$MATCH_RUN_LEN+256*MIN_MATCH_SIZE, %cx
	call	lzsa1len
#ifdef RAW_FORMAT
	jcxz	lzsa1quit	// bail if we hit EOD
#endif

#if !defined(FLAT16OUT) && !defined(FLAT32)
	xchg	AX, SI		// save %si	
	lea	(BX,DI), SI
	pushw	%ds
	movw	%es, %bp
	cmpw	%si, %di
	jnc	lzsa1sameSeg
	pushw	%si
# ifdef ONLY8086
	pushw	%cx
# endif
	shrclw	$4, %si
# ifdef ONLY8086
	popw	%cx
# endif
	lea	-4096(%bp,%si), %bp
	popw	%si
	andw	$0xF, %si
lzsa1sameSeg:
	movw	%bp, %ds
# ifdef FASTFILL
	cmp	$-FASTFILL,BX
	jae	lzsa1fast
# endif
	rep movsb
lzsa1chunkz:
	popw	%ds
#else
# ifdef FASTFILL
	cmp	$-FASTFILL,BX
	jae	lzsa1fast
# endif
	xchg	AX, SI		// save %si	
	lea	(BX,DI), SI
# ifdef ONLY8086
lzsa2movsb:
	movsb	%es:(SI), %es:(DI)	//  NMOS 8088/8086 workaround.
	loop	lzsa2movsb
# else
	rep movsb	%es:(SI), %es:(DI)
# endif
#define lzsa1chunkz lzsa1chunk
#endif
	xchg	AX, SI		// restore %si	
	jmp	lzsa1chunk
#ifdef FASTFILL
lzsa1fast:
# if FASTFILL == 1
#  if !defined(FLAT16OUT) && !defined(FLAT32)
	lodsb
#  else
	movb	%es:(BX,DI), %al
#  endif
	rep stosb
# endif
# if FASTFILL == 2
#  if !defined(FLAT16OUT) && !defined(FLAT32)
	lodsw
#  else
	movw	%es:(BX,DI), %ax
#  endif
	je	lzsa1fastword
	movb	%ah, %al
lzsa1fastword:
	shr	$1, CX
	rep stosw
	jnc	lzsa1chunkz
	stosb
# endif
	jmp	lzsa1chunkz
#endif

lzsa1len:			// get length in %ecx
	andb	%cl, %al
	cbw			// clear %ah
	cmpb	%al, %cl
	jne	lzsa1minNumber	// S=0-6, L=0-14		%cx = %ch + %al   if (%al & %cl != %cl)
	addb	%al, %ch
	lodsb
lzsa1minNumber:
	addb	%ch, %al
	jnc	lzsa1gotNumber  // 0-255			%cx = %ch + %cl + byte   if (%al & %cl == %cl && %ch + %cl + byte < 0x100)
	movb	%al, %ah	// S=256-1791, L=256-4607 or S=256-511, L=256-511
	jne	lzsa1midNumber
	lodsw			// 0-65535			%cx = word   if (%al & %cl == %cl && %ch + %cl + byte == 0x100)
	.byte	0x3C		// mask lodsb with cmpb $0xAC, %al
lzsa1midNumber:
	lodsb			//				%cx = (%ch + %cl + byte)*256 + byte2   if (%al & %cl == %cl && %ch + %cl + byte > 0x100)
lzsa1gotNumber:
	xchgw	%ax, %cx
lzsa1quit:
	ret
