// Lzsa2Decode:
#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define AX	%ax
#define BX	%bx
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define AX	%eax
#define BX	%ebx
#define SI	%esi
#define DI	%edi
#endif

MATCH_RUN_LEN		=	7
LITERALS_RUN_LEN	=	3
MIN_MATCH_SIZE		=	2
MIN_LITERALS_SIZE	=	0

#define PACKED_ONLY			// assume no copy block, optional
//#define PARANOIA			// cover rare cases, optional
 
.macro	shrcl cnt,obj
#ifdef ONLY8086
	movb	\cnt, %cl
	shr	%cl, \obj
#else
	shr	\cnt, \obj
#endif
.endm

#ifdef FLAT16OUT
#define RAW_FORMAT
#endif
lzsa2main:
#ifdef PARANOIA
	cld
#endif
#ifndef RAW_FORMAT
# ifndef NO_LZSA2_HEADER
	lodsw
	cmpw	$0x9E7B, %ax	// magic
	jne	lzsa2main
	lodsb
//	testb	$0x20, %al	// lzsa2
//	je	lzsa2main
# endif
	xorw	%ax, %ax
	cwd			// no nibble stored
	xchgw	%ax, %di
	shrcl	$4, %ax
	jmp	lzsa2blockz	// %di *MUST* be paragraph aligned
# ifndef PACKED_ONLY
lzsa2copy:
	movsb			// handle 64K case
	decw	%cx
	rep	movsb		// copy block
# endif
lzsa2block:			// uncompress chunk
	movw	$0x1000, %ax
lzsa2blockz:	
	movw	%es, %bx
	addw	%ax, %bx
	movw	%bx, %es
# ifndef FLAT16
	movw	%si, %ax
	andw	$0xf, %si
	shrcl	$4, %ax
	movw	%ds, %bx
	addw	%ax, %bx
	movw	%bx, %ds
# endif
	lodsw			// block size
	xchgw	%ax, %cx
	movw	%cx, %bp
	lodsb
# ifndef PACKED_ONLY
	orb	%al, %al
	js	lzsa2copy
	jne	lzsa2full	// 64Kb block
# endif
	jcxz	lzsa2quit	// bail if we hit EOD
lzsa2full:
	addw	%si, %bp
#else
	movb	$0, %dh		// no nibble stored
#endif
#ifdef FLAT32
	orl	$-1, %ebx	// set offset bits 31-16 to 1
#endif
lzsa2chunk:			// uncompress chunk
	lodsb			// get token XYZ|LL|MMM
	pushw	%ax		// keep token
	movw	$LITERALS_RUN_LEN+256*MIN_LITERALS_SIZE, %cx
	shrb	%cl, %al	// shift literals length into place
	call	lzsa2len	// %cl = LITERALS_RUN_LEN
	rep	movsb		// copy %cx literals from %ds:%si to %es:%di
	popw	%ax		// restore token
#ifndef RAW_FORMAT
	cmpw	%bp, %si
	je	lzsa2block	// bail if we hit EOD
#endif
	pushw	%ax		// save token
// Decode XYZ bits to SZC flags
	andb	$0xE0, %al	// filter XYZ
	movb	%al, %ah
	orb	$0xDF, %al
	incw	%ax
	sahf			// ah=SZ.A.P.C
	je	no_nibble_in_offset	// Y=1

	call	getNibble       // get nibble for offset bits 0-3, kill %cl
	sahf
	rclb	$1, %al
	xorb	$0xE1, %al	// set offset bits 7-5 to 1
	sahf
	cbw			// set offset bits 15-8 to 1
// 00Z 5-bit offset: read a nibble for offset bits 1-4 and use the inverted bit Z of the token as bit 0 of the offset.
//                   set bits 5-15 of the offset to 1.
	jns	get_match_length	// 5-bit offset
//10Z 13-bit offset: read a nibble for offset bits 9-12 and use the inverted bit Z for bit 8 of the offset,
//                   then read a byte for offset bits 0-7. set bits 13-15 of the offset to 1.
//                   substract 512 from the offset to get the final value.
	subb	$2, %al		// substract 512, clear C
	jmp	get_match_length_1

getNibble:
	xorb	$0xF0, %dh	// toggle nibble stored flags
	movb	%dh, %al
	jns	gotnibble
	lodsb
	movb	$0xF0, %dh
	orb	%al, %dh
	shrcl	$4, %al
gotnibble:
lzsa2quit:
	ret

no_nibble_in_offset:
	movb	$-1, %ah	// set offset bits 15-8 to 1
	jns	offset_9_bit	// X=0 Y=1
	jc	repeat_match	// rep-match X=1 Y=1 Z=1
//110 16-bit offset: read a byte for offset bits 8-15, then another byte for offset bits 0-7.
	lodsb			// Get 2-byte match offset
get_match_length_1:
	xchgb	%al, %ah

offset_9_bit:	// Z=C
// 01Z 9-bit offset: read a byte for offset bits 0-7 and use the inverted bit Z for bit 8 of the offset.
//                   set bits 9-15 of the offset to 1.
	sbbb	%ch, %ah	// clear bit 8 if Z bit is clear (%ch == 0)
	lodsb			// load match offset bits 0-7
get_match_length:
	xchgw	%ax, %bx	// bx: offset
repeat_match:
//111 repeat offset: reuse the offset value of the previous match command.

	popw	%ax		// restore original token in %al
	movw	$MATCH_RUN_LEN+256*MIN_MATCH_SIZE, %cx
	call	lzsa2len
#if !defined(ENHANCED_FORMAT) && defined(RAW_FORMAT)
	je	lzsa2quit	// bail if we hit EOD in (%si-2)
#endif

#if !defined(FLAT16OUT) && !defined(FLAT32)
	xchg	AX, SI		// save %si	
	lea	(BX,DI), SI
	pushw	%ds
# ifndef RAW_FORMAT
	pushw	%bp		// save end
# endif
	movw	%es, %bp
	cmpw	%si, %di
	jnc	lzsa2sameSeg
	pushw	%si
# ifdef ONLY8086
	pushw	%cx
# endif
	shrcl	$4, %si
# ifdef ONLY8086
	popw	%cx
# endif
	lea	-4096(%bp,%si), %bp
	popw	%si
	andw	$0xF, %si
lzsa2sameSeg:
	movw	%bp, %ds
# ifdef FASTFILL
	cmp	$-FASTFILL,BX
	jae	lzsa2fast
# endif
	rep movsb
lzsa2chunkz:
# ifndef RAW_FORMAT
	popw	%bp		// restore end
# endif
	popw	%ds
#else
# ifdef FASTFILL
	cmp	$-FASTFILL,BX
	jae	lzsa2fast
# endif
	xchg	AX, SI		// save %si	
	lea	(BX,DI), SI
# ifdef ONLY8086
lzsa2movsb:
	movsb	%es:(SI), %es:(DI)	//  NMOS 8088/8086 workaround.
	loop	lzsa2movsb
# else
	rep movsb	%es:(SI), %es:(DI)
# endif
#define lzsa2chunkz lzsa2chunk
#endif
	xchg	AX, SI		// restore %si	
	jmp	lzsa2chunk
#ifdef FASTFILL
lzsa2fast:
# if FASTFILL == 1
#  if !defined(FLAT16OUT) && !defined(FLAT32)
	lodsb
#  else
	movb	%es:(BX,DI), %al
#  endif
	rep stosb
# endif
# if FASTFILL == 2
#  if !defined(FLAT16OUT) && !defined(FLAT32)
	lodsw
#  else
	movw	%es:(BX,DI), %ax
#  endif
	je	lzsa2fastword
	movb	%ah, %al
lzsa2fastword:
	shr	$1, CX
	rep stosw
	jnc	lzsa2chunkz
	stosb
# endif
	jmp	lzsa2chunkz
#endif

lzsa2len:			// get length in %cx
	andb	%cl, %al
	cbw			// clear %ah
	cmpb	%al, %cl
	jne	lzsa2minNumber	// S=0-2, L=0-6			%cx = %ch + %al   if (%al & %cl != %cl)
	addb	%al, %ch
	call	getNibble	// kill %cl
	cmp	$0xF, %al
	jne	lzsa2minNumber	//				%cx = %ch + %cl + nibble   if (%al & %cl == %cl && nibble != 0xF)
	addb	%al, %ch
	lodsb
lzsa2minNumber:
	addb	%ch, %al
	jnc	lzsa2gotNumber  // 0-255			%cx = %ch + %cl + 0xF + byte   if (%al & %cl == %cl && nibble == 0xF && %ch + %cl + 0xF + byte < 0x100)
#ifdef ENHANCED_FORMAT
	decw	%ax
	je	lzsa2maxNumber
# ifdef RAW_FORMAT
	jns	lzsa2midNumber
	popw	%ax
	ret			// bail if we hit EOD
lzsa2midNumber:
# endif
	xchgb	%al, %ah	// S=256-4351, L=256-5887
	lodsb			//				%cx = (%ch + %cl + 0xE + byte)*256 + byte2   if (%al & %cl == %cl && nibble == 0xF && %ch + %cl + 0xF + byte > 0x100)
	.byte	0xB1		// mask lodsw with movb $0xAD, %cl
lzsa2maxNumber:
#endif
	lodsw			// 0-65535			%cx = word   if (%al & %cl == %cl && nibble == 0xF && %ch + %cl + 0xF + byte == 0x100)
lzsa2gotNumber:
	xchgw	%ax, %cx
	ret
